/*
 * For loongson7a dc controller dvo1 i2c
 * Use gpio to access ch7034 chip
 * */

#include <pmon.h>
#include <stdio.h>

#define SCL1	3
#define	SDA1	2

#define SCL0	1
#define	SDA0	0

#define	LOW	0
#define HIGH	1

#define	OUT	0
#define	IN	1

#define CHIP_7034_ADDR	0xea
#define CHIP_9022_ADDR	0x72
#define	ADV7513_ADDR	0x7a
#define	ADV7511_ADDR	0x72
#define CHIP_7053_ADDR	0xec

static unsigned int dc_base = 0;
static unsigned int gpio_val_base = 0;
static unsigned int gpio_dir_base = 0;
static int current_sda, current_scl;

unsigned char CH7034_VGA_REG_TABLE2[][131][2] = {
      {	//mtfvga
      //IN 800x600,out 800x600,ch7034,bypassmode vga,mode_idx=1
          { 0x03, 0x04 },//page 4
          { 0x52, 0xC3 },
          { 0x5A, 0x06 },
          { 0x5A, 0x04 },
          { 0x5A, 0x06 },
          { 0x52, 0xC1 },
          { 0x52, 0xC3 },
          { 0x5A, 0x04 },

          { 0x03, 0x00 },//page 1
          { 0x07, 0xD9 },
          { 0x08, 0xF1 },
          { 0x09, 0x13 },
          { 0x0A, 0xBE },
          { 0x0B, 0x23 },
          { 0x0C, 0x20 },
          { 0x0D, 0x20 },
          { 0x0E, 0x00 },
          { 0x0F, 0x28 },
          { 0x10, 0x80 },
          { 0x11, 0x12 },
          { 0x12, 0x58 },
          { 0x13, 0x74 },
          { 0x14, 0x00 },
          { 0x15, 0x01 },
          { 0x16, 0x04 },
          { 0x17, 0x00 },
          { 0x18, 0x00 },//mtf modify 888RGB
         // { 0x18, 0x20 },//565RGB
//          { 0x18, 0x02 },//888GRB
          { 0x19, 0xF8 },//freq
          { 0x1A, 0x9B },
          { 0x1B, 0x78 },
          { 0x1C, 0x69 },
          { 0x1D, 0x78 },
          { 0x1E, 0x00 },//output is  progressive
          { 0x1F, 0x23 },
          { 0x20, 0x20 },
          { 0x21, 0x20 },
          { 0x22, 0x00 },
          { 0x23, 0x10 },
          { 0x24, 0x60 },
          { 0x25, 0x12 },
          { 0x26, 0x58 },
          { 0x27, 0x74 },
          { 0x28, 0x00 },
          { 0x29, 0x0A },
          { 0x2A, 0x02 },
          { 0x2B, 0x09 },//vga output format:bypass mode
          { 0x2C, 0x00 },
          { 0x2D, 0x00 },
          { 0x2E, 0x3D },
          { 0x2F, 0x00 },//??
          { 0x32, 0xC0 },//???
          { 0x36, 0x40 },
          { 0x38, 0x47 },
          { 0x3D, 0x86 },
          { 0x3E, 0x00 },
          { 0x40, 0x0E },
          { 0x4B, 0x40 },//pwm control
          { 0x4C, 0x40 },//lvds output channel order register
          { 0x4D, 0x80 },
          { 0x54, 0x80 },//lvds
          { 0x55, 0x28 },//lvds
          { 0x56, 0x80 },//lvds
          { 0x57, 0x00 },//lvds
          { 0x58, 0x01 },//lvds
          { 0x59, 0x04 },//lvds
          { 0x5A, 0x02 },
          { 0x5B, 0xF2 },
          { 0x5C, 0xB9 },
          { 0x5D, 0xD6 },
          { 0x5E, 0x54 },
          { 0x60, 0x00 },
          { 0x61, 0x00 },
          { 0x64, 0x2D },
          { 0x68, 0x44 },
          { 0x6A, 0x40 },
          { 0x6B, 0x00 },
          { 0x6C, 0x10 },
          { 0x6D, 0x00 },
          { 0x6E, 0xA0 },
          { 0x70, 0x98 },
          { 0x74, 0x30 },//scaling control
          { 0x75, 0x80 },//scaling control
          { 0x7E, 0x0F },//de and pwm control
          { 0x7F, 0x00 },//test pattern

          { 0x03, 0x01 },//page 2
          { 0x08, 0x05 },
          { 0x09, 0x04 },//diffen register
          { 0x0B, 0x65 },
          { 0x0C, 0x4A },
          { 0x0D, 0x29 },
          { 0x0F, 0x9C },
          { 0x12, 0xD4 },
          { 0x13, 0x28 },
          { 0x14, 0x83 },
          { 0x15, 0x00 },
          { 0x16, 0x00 },
          { 0x1A, 0x6C },//DAC termination control register
          { 0x1B, 0x00 },
          { 0x1C, 0x00 },
          { 0x1D, 0x00 },
          { 0x23, 0x63 },
          { 0x24, 0xB4 },
          { 0x28, 0x54 },
          { 0x29, 0x60 },
          { 0x41, 0x60 },//lvds
          { 0x63, 0x2D },//DE polarity
          { 0x6B, 0x11 },
          { 0x6C, 0x06 },

          { 0x03, 0x03 },//page3
          { 0x26, 0x00 },
          { 0x28, 0x08 },//output control:DAC output is VGA
          { 0x2A, 0x00 },//output control:HDTV output through scaler

          { 0x03, 0x04 },//page 4
          { 0x10, 0x00 },
          { 0x11, 0x9B },
          { 0x12, 0x78 },
          { 0x13, 0x02 },
          { 0x14, 0x88 },
          { 0x15, 0x70 },
          { 0x20, 0x00 },
          { 0x21, 0x00 },
          { 0x22, 0x00 },
          { 0x23, 0x00 },
          { 0x24, 0x00 },
          { 0x25, 0x00 },
          { 0x26, 0x00 },
          { 0x54, 0xC4 },
          { 0x55, 0x5B },
          { 0x56, 0x4D },
          { 0x60, 0x01 },
          { 0x61, 0x62 },
      },
};

unsigned char CH7053_VGA_REG_TABLE[][2] = {
//Page 0 (0x03 = 0x00):
    { 0x00, 0x5E },
    { 0x01, 0x03 },
    { 0x02, 0x33 },
    { 0x03, 0x00 },
    { 0x04, 0x06 },
    { 0x05, 0x58 },
    { 0x06, 0xAC },
    { 0x07, 0xD4 },
    { 0x08, 0x01 },
    { 0x09, 0x1A },
    { 0x0A, 0x28 },
    { 0x0B, 0x47 },
    { 0x0C, 0x80 },
    { 0x0D, 0x98 },
    { 0x0E, 0x00 },
    { 0x0F, 0x58 },
    { 0x10, 0x2C },
    { 0x11, 0x24 },
    { 0x12, 0x38 },
    { 0x13, 0x65 },
    { 0x14, 0x00 },
    { 0x15, 0x04 },
    { 0x16, 0x05 },
    { 0x17, 0x00 },
    { 0x18, 0x00 },
    { 0x19, 0xBA },
    { 0x1A, 0x44 },
    { 0x1B, 0x14 },
    { 0x1C, 0x69 },
    { 0x1D, 0x78 },
    { 0x1E, 0x00 },
    { 0x1F, 0x47 },
    { 0x20, 0x80 },
    { 0x21, 0x98 },
    { 0x22, 0x00 },
    { 0x23, 0x10 },
    { 0x24, 0x60 },
    { 0x25, 0x24 },
    { 0x26, 0x38 },
    { 0x27, 0x65 },
    { 0x28, 0x00 },
    { 0x29, 0x0A },
    { 0x2A, 0x02 },
    { 0x2B, 0x09 },
    { 0x2C, 0x00 },
    { 0x2D, 0x00 },
    { 0x2E, 0x3D },
    { 0x2F, 0x00 },
    { 0x30, 0x01 },
    { 0x31, 0x01 },
    { 0x32, 0xC0 },
    { 0x33, 0x01 },
    { 0x34, 0x01 },
    { 0x35, 0x80 },
    { 0x36, 0x40 },
    { 0x37, 0x40 },
    { 0x38, 0x47 },
    { 0x39, 0x88 },
    { 0x3A, 0x00 },
    { 0x3B, 0x00 },
    { 0x3C, 0x00 },
    { 0x3D, 0x8A },
    { 0x3E, 0x00 },
    { 0x3F, 0x11 },
    { 0x40, 0x04 },
    { 0x41, 0x00 },
    { 0x42, 0x00 },
    { 0x43, 0x00 },
    { 0x44, 0x00 },
    { 0x45, 0x00 },
    { 0x46, 0x00 },
    { 0x47, 0x00 },
    { 0x48, 0x00 },
    { 0x49, 0x00 },
    { 0x4A, 0x00 },
    { 0x4B, 0x40 },
    { 0x4C, 0x40 },
    { 0x4D, 0x80 },
    { 0x4E, 0x00 },
    { 0x4F, 0x00 },
    { 0x50, 0x00 },
    { 0x51, 0x1F },
    { 0x52, 0xFF },
    { 0x53, 0x00 },
    { 0x54, 0x80 },
    { 0x55, 0x58 },
    { 0x56, 0x2C },
    { 0x57, 0x00 },
    { 0x58, 0x04 },
    { 0x59, 0x05 },
    { 0x5A, 0x08 },
    { 0x5B, 0x00 },
    { 0x5C, 0x00 },
    { 0x5D, 0x00 },
    { 0x5E, 0x54 },
    { 0x5F, 0x00 },
    { 0x60, 0x00 },
    { 0x61, 0x00 },
    { 0x62, 0x00 },
    { 0x63, 0x01 },
    { 0x64, 0x2D },
    { 0x65, 0x90 },
    { 0x66, 0x20 },
    { 0x67, 0x22 },
    { 0x68, 0x44 },
    { 0x69, 0x24 },
    { 0x6A, 0x40 },
    { 0x6B, 0x00 },
    { 0x6C, 0x10 },
    { 0x6D, 0x00 },
    { 0x6E, 0xA0 },
    { 0x6F, 0x4B },
    { 0x70, 0x98 },
    { 0x71, 0x01 },
    { 0x72, 0x00 },
    { 0x73, 0x00 },
    { 0x74, 0x20 },
    { 0x75, 0x80 },
    { 0x76, 0x18 },
    { 0x77, 0x00 },
    { 0x78, 0x00 },
    { 0x79, 0x01 },
    { 0x7A, 0x00 },
    { 0x7B, 0x00 },
    { 0x7C, 0x00 },
    { 0x7D, 0xFF },
    { 0x7E, 0x8F },
    { 0x7F, 0x00 },

//Page 1 (0x03 = 0x01):
    { 0x00, 0x56 },
    { 0x01, 0x03 },
    { 0x02, 0x33 },
    { 0x03, 0x01 },
    { 0x04, 0x06 },
    { 0x05, 0x58 },
    { 0x06, 0xAC },
    { 0x07, 0x62 },
    { 0x08, 0x05 },
    { 0x09, 0x04 },
    { 0x0A, 0xC0 },
    { 0x0B, 0x7D },
    { 0x0C, 0x74 },
    { 0x0D, 0x21 },
    { 0x0E, 0x80 },
    { 0x0F, 0x9C },
    { 0x10, 0x00 },
    { 0x11, 0x78 },
    { 0x12, 0xD6 },
    { 0x13, 0x28 },
    { 0x14, 0x80 },
    { 0x15, 0xE1 },
    { 0x16, 0x00 },
    { 0x17, 0x0E },
    { 0x18, 0xC8 },
    { 0x19, 0x42 },
    { 0x1A, 0x6C },
    { 0x1B, 0x00 },
    { 0x1C, 0x00 },
    { 0x1D, 0x00 },
    { 0x1E, 0x09 },
    { 0x1F, 0xA0 },
    { 0x20, 0x00 },
    { 0x21, 0x10 },
    { 0x22, 0x07 },
    { 0x23, 0x63 },
    { 0x24, 0xB4 },
    { 0x25, 0x10 },
    { 0x26, 0x00 },
    { 0x27, 0x00 },
    { 0x28, 0x54 },
    { 0x29, 0x60 },
    { 0x2A, 0x00 },
    { 0x2B, 0x00 },
    { 0x2C, 0x00 },
    { 0x2D, 0x00 },
    { 0x2E, 0x00 },
    { 0x2F, 0x00 },
    { 0x30, 0x00 },
    { 0x31, 0x00 },
    { 0x32, 0x00 },
    { 0x33, 0x00 },
    { 0x34, 0x00 },
    { 0x35, 0x00 },
    { 0x36, 0x00 },
    { 0x37, 0x00 },
    { 0x38, 0x00 },
    { 0x39, 0x00 },
    { 0x3A, 0x00 },
    { 0x3B, 0x0B },
    { 0x3C, 0x00 },
    { 0x3D, 0x00 },
    { 0x3E, 0x00 },
    { 0x3F, 0x00 },
    { 0x40, 0x08 },
    { 0x41, 0x60 },
    { 0x42, 0x14 },
    { 0x43, 0x20 },
    { 0x44, 0x00 },
    { 0x45, 0x00 },
    { 0x46, 0x20 },
    { 0x47, 0x00 },
    { 0x48, 0x49 },
    { 0x49, 0x10 },
    { 0x4A, 0xFF },
    { 0x4B, 0xFF },
    { 0x4C, 0xFF },
    { 0x4D, 0x00 },
    { 0x4E, 0x08 },
    { 0x4F, 0x00 },
    { 0x50, 0x00 },
    { 0x51, 0x00 },
    { 0x52, 0x00 },
    { 0x53, 0x00 },
    { 0x54, 0x00 },
    { 0x55, 0x80 },
    { 0x56, 0x00 },
    { 0x57, 0x00 },
    { 0x58, 0x00 },
    { 0x59, 0x00 },
    { 0x5A, 0x00 },
    { 0x5B, 0x7A },
    { 0x5C, 0x5E },
    { 0x5D, 0x6E },
    { 0x5E, 0x1F },
    { 0x5F, 0x1F },
    { 0x60, 0x00 },
    { 0x61, 0x00 },
    { 0x62, 0x00 },
    { 0x63, 0x2D },
    { 0x64, 0x40 },
    { 0x65, 0x40 },
    { 0x66, 0x40 },
    { 0x67, 0x00 },
    { 0x68, 0x00 },
    { 0x69, 0x00 },
    { 0x6A, 0x00 },
    { 0x6B, 0x10 },
    { 0x6C, 0x00 },
    { 0x6D, 0xA9 },
    { 0x6E, 0x90 },
    { 0x6F, 0x50 },
    { 0x70, 0x00 },
    { 0x71, 0x00 },
    { 0x72, 0x09 },
    { 0x73, 0x00 },
    { 0x74, 0x00 },
    { 0x75, 0x70 },
    { 0x76, 0x00 },
    { 0x77, 0x50 },
    { 0x78, 0x00 },
    { 0x79, 0x98 },
    { 0x7A, 0x00 },
    { 0x7B, 0x98 },
    { 0x7C, 0x03 },
    { 0x7D, 0x00 },
    { 0x7E, 0x00 },
    { 0x7F, 0x00 },

//Page 2 (0x03 = 0x02):
    { 0x00, 0x5E },
    { 0x01, 0x03 },
    { 0x02, 0x33 },
    { 0x03, 0xFF },
    { 0x04, 0x06 },
    { 0x05, 0x58 },
    { 0x06, 0xAC },
    { 0x07, 0xFF },
    { 0x08, 0xFF },
    { 0x09, 0xFF },
    { 0x0A, 0xFF },
    { 0x0B, 0xFF },
    { 0x0C, 0xFF },
    { 0x0D, 0xFF },
    { 0x0E, 0xFF },
    { 0x0F, 0xFF },
    { 0x10, 0xFF },
    { 0x11, 0xFF },
    { 0x12, 0xFF },
    { 0x13, 0xFF },
    { 0x14, 0xFF },
    { 0x15, 0xFF },
    { 0x16, 0xFF },
    { 0x17, 0xFF },
    { 0x18, 0xFF },
    { 0x19, 0xFF },
    { 0x1A, 0xFF },
    { 0x1B, 0xFF },
    { 0x1C, 0xFF },
    { 0x1D, 0xFF },
    { 0x1E, 0xFF },
    { 0x1F, 0xFF },
    { 0x20, 0xFF },
    { 0x21, 0xFF },
    { 0x22, 0xFF },
    { 0x23, 0xFF },
    { 0x24, 0xFF },
    { 0x25, 0xFF },
    { 0x26, 0xFF },
    { 0x27, 0xFF },
    { 0x28, 0xFF },
    { 0x29, 0xFF },
    { 0x2A, 0xFF },
    { 0x2B, 0xFF },
    { 0x2C, 0xFF },
    { 0x2D, 0xFF },
    { 0x2E, 0xFF },
    { 0x2F, 0xFF },
    { 0x30, 0xFF },
    { 0x31, 0xFF },
    { 0x32, 0xFF },
    { 0x33, 0xFF },
    { 0x34, 0xFF },
    { 0x35, 0xFF },
    { 0x36, 0xFF },
    { 0x37, 0xFF },
    { 0x38, 0xFF },
    { 0x39, 0xFF },
    { 0x3A, 0xFF },
    { 0x3B, 0xFF },
    { 0x3C, 0xFF },
    { 0x3D, 0xFF },
    { 0x3E, 0xFF },
    { 0x3F, 0xFF },
    { 0x40, 0xFF },
    { 0x41, 0xFF },
    { 0x42, 0xFF },
    { 0x43, 0xFF },
    { 0x44, 0xFF },
    { 0x45, 0xFF },
    { 0x46, 0xFF },
    { 0x47, 0xFF },
    { 0x48, 0xFF },
    { 0x49, 0xFF },
    { 0x4A, 0xFF },
    { 0x4B, 0xFF },
    { 0x4C, 0xFF },
    { 0x4D, 0xFF },
    { 0x4E, 0xFF },
    { 0x4F, 0xFF },
    { 0x50, 0xFF },
    { 0x51, 0xFF },
    { 0x52, 0xFF },
    { 0x53, 0xFF },
    { 0x54, 0xFF },
    { 0x55, 0xFF },
    { 0x56, 0xFF },
    { 0x57, 0xFF },
    { 0x58, 0xFF },
    { 0x59, 0xFF },
    { 0x5A, 0xFF },
    { 0x5B, 0xFF },
    { 0x5C, 0xFF },
    { 0x5D, 0xFF },
    { 0x5E, 0xFF },
    { 0x5F, 0xFF },
    { 0x60, 0xFF },
    { 0x61, 0xFF },
    { 0x62, 0xFF },
    { 0x63, 0xFF },
    { 0x64, 0xFF },
    { 0x65, 0xFF },
    { 0x66, 0xFF },
    { 0x67, 0xFF },
    { 0x68, 0xFF },
    { 0x69, 0xFF },
    { 0x6A, 0xFF },
    { 0x6B, 0xFF },
    { 0x6C, 0xFF },
    { 0x6D, 0xFF },
    { 0x6E, 0xFF },
    { 0x6F, 0xFF },
    { 0x70, 0xFF },
    { 0x71, 0xFF },
    { 0x72, 0xFF },
    { 0x73, 0xFF },
    { 0x74, 0xFF },
    { 0x75, 0xFF },
    { 0x76, 0xFF },
    { 0x77, 0xFF },
    { 0x78, 0xFF },
    { 0x79, 0xFF },
    { 0x7A, 0xFF },
    { 0x7B, 0xFF },
    { 0x7C, 0xFF },
    { 0x7D, 0xFF },
    { 0x7E, 0xFF },
    { 0x7F, 0xFF },

//Page 3 (0x03 = 0x03):
    { 0x00, 0x5E },
    { 0x01, 0x03 },
    { 0x02, 0x33 },
    { 0x03, 0x03 },
    { 0x04, 0x06 },
    { 0x05, 0x58 },
    { 0x06, 0xAC },
    { 0x07, 0x00 },
    { 0x08, 0x00 },
    { 0x09, 0x00 },
    { 0x0A, 0x00 },
    { 0x0B, 0x00 },
    { 0x0C, 0xFF },
    { 0x0D, 0x00 },
    { 0x0E, 0x03 },
    { 0x0F, 0x19 },
    { 0x10, 0x40 },
    { 0x11, 0x01 },
    { 0x12, 0x00 },
    { 0x13, 0x00 },
    { 0x14, 0x00 },
    { 0x15, 0x00 },
    { 0x16, 0x00 },
    { 0x17, 0x00 },
    { 0x18, 0x00 },
    { 0x19, 0x00 },
    { 0x1A, 0x00 },
    { 0x1B, 0x00 },
    { 0x1C, 0x00 },
    { 0x1D, 0x00 },
    { 0x1E, 0x00 },
    { 0x1F, 0x00 },
    { 0x20, 0x00 },
    { 0x21, 0x00 },
    { 0x22, 0x29 },
    { 0x23, 0xD5 },
    { 0x24, 0x89 },
    { 0x25, 0x11 },
    { 0x26, 0x00 },
    { 0x27, 0x01 },
    { 0x28, 0x0C },
    { 0x29, 0x00 },
    { 0x2A, 0x28 },
    { 0x2B, 0x00 },
    { 0x2C, 0x00 },
    { 0x2D, 0x00 },
    { 0x2E, 0x80 },
    { 0x2F, 0x80 },
    { 0x30, 0x00 },
    { 0x31, 0x00 },
    { 0x32, 0x00 },
    { 0x33, 0xFF },
    { 0x34, 0xFF },
    { 0x35, 0xFF },
    { 0x36, 0xFF },
    { 0x37, 0xFF },
    { 0x38, 0xFF },
    { 0x39, 0xFF },
    { 0x3A, 0xFF },
    { 0x3B, 0xFF },
    { 0x3C, 0xFF },
    { 0x3D, 0xFF },
    { 0x3E, 0xFF },
    { 0x3F, 0xFF },
    { 0x40, 0xFF },
    { 0x41, 0x09 },
    { 0x42, 0x1D },
    { 0x43, 0x0F },
    { 0x44, 0x00 },
    { 0x45, 0xFF },
    { 0x46, 0xFF },
    { 0x47, 0xFF },
    { 0x48, 0xFF },
    { 0x49, 0xFF },
    { 0x4A, 0xFF },
    { 0x4B, 0xFF },
    { 0x4C, 0xFF },
    { 0x4D, 0xFF },
    { 0x4E, 0xFF },
    { 0x4F, 0xFF },
    { 0x50, 0xFF },
    { 0x51, 0xFF },
    { 0x52, 0xFF },
    { 0x53, 0xFF },
    { 0x54, 0xFF },
    { 0x55, 0xFF },
    { 0x56, 0xFF },
    { 0x57, 0xFF },
    { 0x58, 0xFF },
    { 0x59, 0xFF },
    { 0x5A, 0xFF },
    { 0x5B, 0xFF },
    { 0x5C, 0xFF },
    { 0x5D, 0xFF },
    { 0x5E, 0xFF },
    { 0x5F, 0xFF },
    { 0x60, 0xFF },
    { 0x61, 0xFF },
    { 0x62, 0xFF },
    { 0x63, 0xFF },
    { 0x64, 0xFF },
    { 0x65, 0xFF },
    { 0x66, 0xFF },
    { 0x67, 0xFF },
    { 0x68, 0xFF },
    { 0x69, 0xFF },
    { 0x6A, 0xFF },
    { 0x6B, 0xFF },
    { 0x6C, 0xFF },
    { 0x6D, 0xFF },
    { 0x6E, 0x00 },
    { 0x6F, 0x00 },
    { 0x70, 0xFF },
    { 0x71, 0xF8 },
    { 0x72, 0xFF },
    { 0x73, 0x00 },
    { 0x74, 0x02 },
    { 0x75, 0x00 },
    { 0x76, 0x00 },
    { 0x77, 0x00 },
    { 0x78, 0x00 },
    { 0x79, 0xFF },
    { 0x7A, 0xFF },
    { 0x7B, 0xFF },
    { 0x7C, 0xFF },
    { 0x7D, 0xFF },
    { 0x7E, 0xFF },
    { 0x7F, 0xFF },

//Page 4 (0x03 = 0x04):
    { 0x00, 0x5E },
    { 0x01, 0x03 },
    { 0x02, 0x33 },
    { 0x03, 0x04 },
    { 0x04, 0x06 },
    { 0x05, 0x58 },
    { 0x06, 0xAC },
    { 0x07, 0xFF },
    { 0x08, 0x00 },
    { 0x09, 0x00 },
    { 0x0A, 0x00 },
    { 0x0B, 0x00 },
    { 0x0C, 0x00 },
    { 0x0D, 0x00 },
    { 0x0E, 0x00 },
    { 0x0F, 0x00 },
    { 0x10, 0x02 },
    { 0x11, 0x44 },
    { 0x12, 0x14 },
    { 0x13, 0x02 },
    { 0x14, 0x88 },
    { 0x15, 0x70 },
    { 0x16, 0x00 },
    { 0x17, 0x00 },
    { 0x18, 0x00 },
    { 0x19, 0x00 },
    { 0x1A, 0xFF },
    { 0x1B, 0xFF },
    { 0x1C, 0xFF },
    { 0x1D, 0xFF },
    { 0x1E, 0xFF },
    { 0x1F, 0xFF },
    { 0x20, 0x00 },
    { 0x21, 0x00 },
    { 0x22, 0x00 },
    { 0x23, 0x00 },
    { 0x24, 0x00 },
    { 0x25, 0x00 },
    { 0x26, 0x00 },
    { 0x27, 0x00 },
    { 0x28, 0x00 },
    { 0x29, 0x00 },
    { 0x2A, 0xEF },
    { 0x2B, 0x07 },
    { 0x2C, 0xEF },
    { 0x2D, 0x07 },
    { 0x2E, 0x86 },
    { 0x2F, 0x07 },
    { 0x30, 0x86 },
    { 0x31, 0x07 },
    { 0x32, 0xF0 },
    { 0x33, 0x00 },
    { 0x34, 0xF0 },
    { 0x35, 0x00 },
    { 0x36, 0x10 },
    { 0x37, 0x00 },
    { 0x38, 0x00 },
    { 0x39, 0x10 },
    { 0x3A, 0x00 },
    { 0x3B, 0x00 },
    { 0x3C, 0x00 },
    { 0x3D, 0x00 },
    { 0x3E, 0x00 },
    { 0x3F, 0x00 },
    { 0x40, 0x00 },
    { 0x41, 0xFF },
    { 0x42, 0xFF },
    { 0x43, 0xFF },
    { 0x44, 0xFF },
    { 0x45, 0x00 },
    { 0x46, 0x03 },
    { 0x47, 0x00 },
    { 0x48, 0x00 },
    { 0x49, 0x00 },
    { 0x4A, 0x24 },
    { 0x4B, 0x00 },
    { 0x4C, 0x48 },
    { 0x4D, 0xFF },
    { 0x4E, 0xFF },
    { 0x4F, 0x7F },
    { 0x50, 0x5E },
    { 0x51, 0xC3 },
    { 0x52, 0xFB },
    { 0x53, 0xFF },
    { 0x54, 0xD4 },
    { 0x55, 0x60 },
    { 0x56, 0xAD },
    { 0x57, 0x48 },
    { 0x58, 0xFF },
    { 0x59, 0xFF },
    { 0x5A, 0xFF },
    { 0x5B, 0x00 },
    { 0x5C, 0xFF },
    { 0x5D, 0xFF },
    { 0x5E, 0xFF },
    { 0x5F, 0xFF },
    { 0x60, 0x01 },
    { 0x61, 0x62 },
    { 0x62, 0xFF },
    { 0x63, 0xFF },
    { 0x64, 0xFF },
    { 0x65, 0xFF },
    { 0x66, 0xFF },
    { 0x67, 0xFF },
    { 0x68, 0xFF },
    { 0x69, 0xFF },
    { 0x6A, 0xFF },
    { 0x6B, 0xFF },
    { 0x6C, 0xFF },
    { 0x6D, 0xFF },
    { 0x6E, 0xFF },
    { 0x6F, 0xFF },
    { 0x70, 0xFF },
    { 0x71, 0xFF },
    { 0x72, 0xFF },
    { 0x73, 0xFF },
    { 0x74, 0xFF },
    { 0x75, 0xFF },
    { 0x76, 0xFF },
    { 0x77, 0xFF },
    { 0x78, 0xFF },
    { 0x79, 0xFF },
    { 0x7A, 0xFF },
    { 0x7B, 0xFF },
    { 0x7C, 0xFF },
    { 0x7D, 0xFF },
    { 0x7E, 0xFF },
    { 0x7F, 0xFF },
};

#define CH7053_REGMAP_LENGTH_VGA (sizeof(CH7053_VGA_REG_TABLE) / (2*sizeof(CH7053_VGA_REG_TABLE[0][0])))
#define CH7034_REGMAP_LENGTH_VGA (sizeof(CH7034_VGA_REG_TABLE2[0]) / (2*sizeof(unsigned char)))

void gpioi2c_init()
{
	dc_base = *(volatile unsigned int *)0xba003110;
	dc_base &= 0xfffffff0;
	dc_base |= 0x80000000;

	gpio_val_base = dc_base + 0x1650;
	gpio_dir_base = dc_base + 0x1660;
#if 0
	printf("gpioi2c init regs base\n");
	printf("gpio_val_base 0x%x\n", gpio_val_base);
	printf("gpio_dir_base 0x%x\n", gpio_dir_base);
#endif
}

static void set_gpio_value(int gpio, int val)
{
	if (val == 1)
		*(volatile unsigned int *)gpio_val_base |= (1U << gpio);
	else
		*(volatile unsigned int *)gpio_val_base &= ~(1U << gpio);
}

static unsigned int get_gpio_value(int gpio)
{
	return (*(volatile unsigned int *)gpio_val_base & (1U << gpio)) ? 1 : 0;
}

static void set_gpio_direction(unsigned int gpio, int val)
{
	if (val == 1)
		*(volatile unsigned int *)gpio_dir_base |= (1 << gpio); 
	else
		*(volatile unsigned int *)gpio_dir_base &= ~(1 << gpio); 
}

static int gpioi2c_read_ack(int scl, int sda)
{
	set_gpio_direction(sda, IN);
	delay(200);

	set_gpio_value(scl, HIGH);
	delay(200);

	/* the ack signal will hold on sda untill scl falling edge */
	while(get_gpio_value(sda))
		delay(200);

	set_gpio_value(scl, LOW);
	delay(2000);

	return 0;
}

#if 0
static void gpioi2c_send_ack()
{
	set_gpio_direction(SDA, OUT);
	set_gpio_value(SCL, LOW);
	delay(200);
	set_gpio_value(SDA, 0);
	delay(200);
	set_gpio_value(SCL, HIGH);
	delay(200);
}
#endif

static int gpioi2c_write_byte(unsigned char dev_addr, unsigned char c)
{
	int i;
	int sda, scl;

	if (dev_addr == CHIP_7034_ADDR) {
		sda = SDA1;
		scl = SCL1;
	} else if (dev_addr == CHIP_9022_ADDR) {
		sda = SDA0;
		scl = SCL0;
	}
	else  {
		sda = current_sda;
		scl = current_scl;
	}

	set_gpio_direction(sda, OUT);

	for(i = 7; i >= 0; i--)
	{
		set_gpio_value(scl, LOW);
		delay(200);
		set_gpio_value(sda, (c & (1U << i)) ? 1 : 0);//high bit --> low bit
		delay(200);
		set_gpio_value(scl, HIGH);
		delay(200);
	}

	set_gpio_value(scl, LOW);
	delay(200);

	if (gpioi2c_read_ack(scl, sda))
	{
		printf("read slave dev ack invalid\n");//0:valid
		return 0;
	}
	return 1;
	
}

static void gpioi2c_read_byte(unsigned char dev_addr, unsigned char *c)
{
	int i;
	int sda, scl;

	if (dev_addr == CHIP_7034_ADDR) {
		sda = SDA1;
		scl = SCL1;
	} else if (dev_addr == CHIP_9022_ADDR) {
		sda = SDA0;
		scl = SCL0;
	}
	else  {
		sda = current_sda;
		scl = current_scl;
	}

	*c = 0;
	set_gpio_direction(sda, IN);

	for (i = 7; i >= 0; i--)
	{
		set_gpio_value(scl, HIGH);
		delay(200);
		*c = (*c << 1) | get_gpio_value(sda);
		set_gpio_value(scl, LOW);
		delay(200);
	}
//	gpioi2c_send_ack();
}

static void gpioi2c_start(unsigned char dev_addr)
{
	int sda, scl;

	if (dev_addr == CHIP_7034_ADDR) {
		sda = SDA1;
		scl = SCL1;
	} else if (dev_addr == CHIP_9022_ADDR) {
		sda = SDA0;
		scl = SCL0;
	}
	else  {
		sda = current_sda;
		scl = current_scl;
	}

	/* if set sda output without setting sda high,
 	 * the sda output value may be low
 	 * */
	set_gpio_value(sda, HIGH);

	set_gpio_direction(sda, OUT);
	set_gpio_direction(scl, OUT);

	set_gpio_value(scl, HIGH);
	delay(200);

	/* start signal: sda from high to low */
	set_gpio_value(sda, HIGH);
	delay(200);
	set_gpio_value(sda, LOW);
	delay(200);
}

static void gpioi2c_stop(unsigned char dev_addr)
{
	int sda, scl;

	if (dev_addr == CHIP_7034_ADDR) {
		sda = SDA1;
		scl = SCL1;
	} else if (dev_addr == CHIP_9022_ADDR) {
		sda = SDA0;
		scl = SCL0;
	}
	else  {
		sda = current_sda;
		scl = current_scl;
	}

	set_gpio_direction(sda, OUT);

	set_gpio_value(scl, HIGH);
	delay(200);

	set_gpio_value(sda, LOW);
	delay(200);
	set_gpio_value(sda, HIGH);
	delay(200);
}

static void gpioi2c_write(unsigned char dev_addr, unsigned char data_addr, unsigned char data)
{
	gpioi2c_start(dev_addr);
	if (!gpioi2c_write_byte(dev_addr, dev_addr))
	{
		printf("gpioi2c write dev_addr fail\n");
		return;
	}
	if (!gpioi2c_write_byte(dev_addr, data_addr))
	{
		printf("gpioi2c write data_addr fail\n");
		return;
	}
	if (!gpioi2c_write_byte(dev_addr, data))
	{
		printf("gpioi2c write data fail\n");
		return;
	}
	gpioi2c_stop(dev_addr);
}

static void dvo1_gpioi2c_write(int ac, unsigned char *av[])
{
	unsigned char dev_addr = CHIP_7034_ADDR;
	unsigned char data_addr;
	unsigned char data;

	switch (ac) {
	case 3:
		if (gpio_val_base == 0)
			gpioi2c_init();

		data_addr = strtoul(av[1], NULL, 0);
		data = strtoul(av[2], NULL, 0);

		gpioi2c_write(dev_addr, data_addr, data);
		printf("0x%2x <= 0x%2x \n", data_addr, data);

		break;
	default:
		printf("need 3 param: gw1 3 4");
		break;
	}
}

static void gpioi2c_read(unsigned char dev_addr, unsigned char data_addr, unsigned char *data)
{
	/* bit0 :1 read, 0 write */

	gpioi2c_start(dev_addr);

	if (!gpioi2c_write_byte(dev_addr, dev_addr))
	{
		printf("gpioi2c write dev_addr fail\n");
		return;
	}

	if (!gpioi2c_write_byte(dev_addr, data_addr))
	{
		printf("gpioi2c write data_addr fail\n");
		return;
	}

	gpioi2c_start(dev_addr);
	if (!gpioi2c_write_byte(dev_addr, dev_addr | 0x01))//for read
	{
		printf("gpioi2c write dev_addr fail\n");
		return;
	}

	gpioi2c_read_byte(dev_addr, data);

	gpioi2c_stop(dev_addr);
}

static void dvo1_gpioi2c_read(int ac, unsigned char *av[])
{
	/* bit0 :1 read, 0 write */
	unsigned char dev_addr = CHIP_7034_ADDR;
	unsigned char data_addr;
	unsigned char ret_val;

	switch (ac) {
	case 2:
		if (gpio_val_base == 0)
			gpioi2c_init();

		data_addr = strtoul(av[1], NULL, 0);
		gpioi2c_read(dev_addr, data_addr, &ret_val);
		printf("0x%2x : 0x%2x\n", data_addr, ret_val);
		break;
	default:
		printf("need 2 param: gr1 0");
		break;
	}
}

void gpioi2c_config_adv7053(unsigned char dvo)
{
	int count;
	unsigned char dev_addr;
	unsigned char data_addr;
	unsigned char data;
	unsigned char data2;
    printf("config DVO%d...\n",dvo);

    dev_addr = CHIP_7053_ADDR;
    
	/* get reg base addr */
	gpioi2c_init();

	if(dvo)
	{
	    current_sda = SDA1;
	    current_scl = SCL1;
	}
	else
	{
	    current_sda = SDA0;
	    current_scl = SCL0;
	}
	
	for (count=0;count<CH7053_REGMAP_LENGTH_VGA;count++)
	{
		data_addr = CH7053_VGA_REG_TABLE[count][0];
		data = CH7053_VGA_REG_TABLE[count][1];
#if 0
        if(data_addr == 0x3)
            printf("------------------page : 0x%x--------------\n", data);
        gpioi2c_read(dev_addr, data_addr, &data2);
        if(data2 == data)
            continue;
//         else
            //printf("not eq, data_addr = 0x%x, data 0x%2x, data2 0x%2x\n",data_addr, data, data2);
#endif
		gpioi2c_write(dev_addr, data_addr, data);
#if 0
		gpioi2c_read(dev_addr, data_addr, &data2);

		if (data != data2)
			printf("not eq, data 0x%2x, data2 0x%2x\n\n\n", data, data2);
	    else
	        printf("read ok! data: 0x%x, data2: 0x%x\n\n\n", data, data2);
#endif
	}
}


void gpioi2c_config_ch7034()
{
	int count;
	unsigned char dev_addr = CHIP_7034_ADDR;
	unsigned char data_addr;
	unsigned char data;
	unsigned char data2;

	/* get reg base addr */
	gpioi2c_init();

	for (count = 0; count < CH7034_REGMAP_LENGTH_VGA; count++)
	{
		data_addr = CH7034_VGA_REG_TABLE2[0][count][0];
		data = CH7034_VGA_REG_TABLE2[0][count][1];

		gpioi2c_write(dev_addr, data_addr, data);
#if DEBUG_CH7034
		gpioi2c_read(dev_addr, data_addr, &data2);

		if (data != data2)
			printf("not eq, data 0x%2x, data2 0x%2x\n\n\n", data, data2);
#endif
	}
}

void gpioi2c_config_sii9022a(void)
{
	unsigned char id0;
	unsigned char id1;
	unsigned char id2;
	unsigned char data;

	unsigned char dev_addr = CHIP_9022_ADDR;

	gpioi2c_init();

	gpioi2c_write(dev_addr, 0xc7, 0x00);
	gpioi2c_read(dev_addr, 0x1b, &id0);
	gpioi2c_read(dev_addr, 0x1c, &id1);
	gpioi2c_read(dev_addr, 0x1d, &id2);

	if (id0 != 0xb0 || id1 != 0x2 || id2 != 0x3) {
		printf("id err\n");
		return;
	}
	
	gpioi2c_read(dev_addr, 0x1e, &data);
	data &= ~(0x3);
	gpioi2c_write(dev_addr, 0x1e, data);

	gpioi2c_read(dev_addr, 0x1a, &data);
	data &= ~(1 << 4);
	gpioi2c_write(dev_addr, 0x1a, data);
}

#ifdef USE_ADV7511
void dvo_hdmi_init()
{
	unsigned char dev_addr = CHIP_9022_ADDR;
	unsigned char d;

	gpioi2c_init();

	gpioi2c_write(dev_addr, 0x41, 0x10);
	gpioi2c_read(dev_addr, 0x41, &d); printf("0x41:0x%x\n",d);
	gpioi2c_write(dev_addr, 0x98, 0x3);
	gpioi2c_read(dev_addr, 0x98, &d); printf("0x98:0x%x\n",d);
	gpioi2c_write(dev_addr, 0x9a, 0xe0);
	gpioi2c_write(dev_addr, 0x9c, 0x30);
	gpioi2c_write(dev_addr, 0x9d, 0x61);
	gpioi2c_write(dev_addr, 0xa2, 0xa4);
	gpioi2c_write(dev_addr, 0xa3, 0xa4);
	gpioi2c_write(dev_addr, 0xe0, 0xd0);
	gpioi2c_write(dev_addr, 0xf9, 0x0);
	gpioi2c_write(dev_addr, 0x41, 0x10);

}
#endif



static const Cmd Cmds[] = {
	{"Misc"},
	{"gr1", "", NULL, "read a data from a CH7034 chip", dvo1_gpioi2c_read, 1, 5, 0},
	{"gw1", "", NULL, "write a data to a CH704 chip", dvo1_gpioi2c_write, 1, 5, 0},
	{"cfg", "", NULL, "write stream data to a CH7034 chip", gpioi2c_config_ch7034, 1, 5, 0},
	{0, 0}
};

static void init_cmd __P((void)) __attribute__ ((constructor));
static void init_cmd()
{
	cmdlist_expand(Cmds, 1);
}
